import contextlib
import dataclasses
import importlib.metadata
import os
import pathlib
import typing as t
import warnings
from collections.abc import MutableSequence

import ase
import h5py
import numpy as np

from znh5md.interface.read import getitem
from znh5md.interface.write import extend
from znh5md.misc import open_file

__version__ = importlib.metadata.version("znh5md")


@dataclasses.dataclass
class IO(MutableSequence):
    """A class for handling H5MD files for ASE Atoms objects."""

    filename: str | os.PathLike | None = None
    file_handle: h5py.File | None = None
    pbc_group: bool = True  # Specify PBC per step (Not H5MD conform)
    save_units: bool = True  # Export ASE units into the H5MD file
    author: str = "N/A"
    author_email: str = "N/A"
    creator: str = "znh5md"
    creator_version: str = __version__
    particles_group: str | None = None
    compression: str | None = "gzip"
    compression_opts: int | None = None
    timestep: float = 1.0
    store: t.Literal["time", "linear"] = "linear"
    tqdm_limit: int = 100
    chunk_size: int | None = 64
    use_ase_calc: bool = True

    _store_ase_origin: bool = True  # for testing purposes only

    def __post_init__(self):
        if self.filename is None and self.file_handle is None:
            raise ValueError("Either filename or file_handle must be provided")
        if self.filename is not None and self.file_handle is not None:
            raise ValueError("Only one of filename or file_handle can be provided")
        if self.filename is not None:
            self.filename = pathlib.Path(self.filename)
        self._set_particle_group()
        self._read_author_creator()

    def _set_particle_group(self):
        if self.particles_group is not None:
            pass
        elif self.filename is not None and self.filename.exists():
            with open_file(self.filename, self.file_handle, mode="r") as f:
                self.particles_group = next(iter(f["particles"].keys()))
        elif (
            self.file_handle is not None
            and pathlib.Path(self.file_handle.filename).exists()
        ):
            with open_file(self.filename, self.file_handle, mode="r") as f:
                self.particles_group = next(iter(f["particles"].keys()))
        else:
            self.particles_group = "atoms"  # Default group name

    def _read_author_creator(self):
        with contextlib.suppress(FileNotFoundError, KeyError):
            # FileNotFoundError if the filename does not exist
            # KeyError if the file has not yet been initialized as H5MD
            #   or the keys are not provided, which is officially
            #   not allowed in H5MD.
            with open_file(self.filename, self.file_handle, mode="r") as f:
                self.author = f["h5md"]["author"].attrs["name"]
                self.author_email = f["h5md"]["author"].attrs["email"]
                self.creator = f["h5md"]["creator"].attrs["name"]
                self.creator_version = f["h5md"]["creator"].attrs["version"]

    def create_file(self):
        with open_file(self.filename, self.file_handle, mode="w") as f:
            g_h5md = f.create_group("h5md")
            g_h5md.attrs["version"] = np.array([1, 1])
            g_author = g_h5md.create_group("author")
            g_author.attrs["name"] = self.author
            g_author.attrs["email"] = self.author_email
            g_creator = g_h5md.create_group("creator")
            g_creator.attrs["name"] = self.creator
            g_creator.attrs["version"] = self.creator_version
            f.create_group("particles")

    def __len__(self) -> int:
        with open_file(self.filename, self.file_handle, mode="r") as f:
            return len(f["particles"][self.particles_group]["species"]["value"])

    def __getitem__(
        self, index: int | np.int_ | slice | np.ndarray
    ) -> ase.Atoms | list[ase.Atoms]:
        try:
            return getitem(self, index)
        except FileNotFoundError:
            # FileNotFoundError is an OSError, but we want to handle it
            #  separately from the OSError h5py raises
            raise
        except OSError:
            raise IndexError("Index out of range")

    def extend(self, frames: list[ase.Atoms]) -> None:
        if not isinstance(frames, list):
            raise ValueError("images must be a list of ASE Atoms objects")
        if len(frames) == 0:
            warnings.warn("No data provided")
            return
        if self.filename is not None and not self.filename.exists():
            self.create_file()
        if self.file_handle is not None:
            needs_creation = False
            with open_file(self.filename, self.file_handle, mode="r") as f:
                needs_creation = "h5md" not in f
            if needs_creation:
                self.create_file()

        extend(self, frames)

    def append(self, atoms: ase.Atoms):
        if not isinstance(atoms, ase.Atoms):
            raise ValueError("atoms must be an ASE Atoms object")
        self.extend([atoms])

    def __delitem__(self, index):
        raise NotImplementedError("Deleting items is not supported")

    def __setitem__(self, index, value):
        raise NotImplementedError("Setting items is not supported")

    def insert(self, index, value):
        raise NotImplementedError("Inserting items is not supported")



import contextlib
import os
import typing as t

import h5py
import numpy as np


def concatenate_varying_shape_arrays(
    values: list, fillvalue: str | int | float | bool, dtype
) -> np.ndarray:
    """
    Concatenates arrays of varying shapes into a single array,
    padding smaller arrays with a specified fillvalue.

    Args:
        values (list): List of numpy arrays with varying shapes.
        fillvalue (str | int | float | bool): Value used to fill missing entries.
            The fillvalue also determines the dtype of the output array.

    Returns:
        np.ndarray: A new array containing the input arrays,
        padded to match the maximum shape.
    """

    # Determine the maximum shape along all dimensions
    maxshape = list(values[0].shape)
    for value in values[1:]:
        maxshape = [max(a, b) for a, b in zip(maxshape, value.shape)]

    # Add the batch dimension
    maxshape = (len(values), *maxshape)

    # Create an array filled with the fillvalue
    dataset = np.full(maxshape, fillvalue, dtype=dtype)

    # Insert each value into the dataset
    for i, value in enumerate(values):
        # Create slices for each dimension of the current value
        slices = tuple(slice(0, dim) for dim in value.shape)
        dataset[(i,) + slices] = value

    return dataset


def decompose_varying_shape_arrays(
    dataset: np.ndarray, fillvalue: str | int | float | bool | np.ndarray
) -> list:
    """
    Decomposes a concatenated array with padding into a list of original arrays.

    Parameters
    ----------
        dataset: np.ndarray
            The concatenated array with padding.
        fillvalue: (str | int | float | bool | np.ndarray)
            Value used to fill missing entries in the original concatenation.

    Returns
    -------
        list: List of numpy arrays with the padding removed.
    """
    decomposed = []
    is_nan = np.isnan(fillvalue) if isinstance(fillvalue, float) else None

    for value in dataset:
        slices = []
        # Collapse all other dimensions to find non-fillvalue
        # regions along the current axis
        if is_nan:
            mask = ~np.isnan(value)
        else:
            mask = value != fillvalue
        for axis in range(value.ndim):
            # Sum along all dimensions except the current one
            axis_sum = mask.any(axis=tuple(i for i in range(value.ndim) if i != axis))
            end = len(axis_sum) - np.argmax(
                axis_sum[::-1]
            )  # Last non-fillvalue index + 1
            slices.append(slice(0, end))

        # Use slices to extract the non-fillvalue portion of the array
        decomposed.append(value[tuple(slices)])

    return decomposed


@contextlib.contextmanager
def open_file(
    filename: str | os.PathLike | None, file_handle: h5py.File | None, **kwargs
) -> t.Generator[h5py.File, None, None]:
    if file_handle is not None:
        yield file_handle
    else:
        with h5py.File(filename, **kwargs) as f:
            yield f


def fill_dataset(dataset, new_data, shift, fill_value):
    # shift is applied along axis 0:
    #  a dataset might not have been extenden in the last step
    #  because no data was added.
    #  with the shift we ensure that the missing data along axis 0 is filled with np.nan
    # Axis 0 is the configuration axis
    # Axis 1 is the number of particles axis
    # all following axis are optional, e.g 2 can be (x, y, z) or 1 can already be energy

    old_shape = dataset.shape
    new_shape = new_data.shape

    if len(old_shape) == 1 and len(new_shape) == 1:
        dataset.resize((old_shape[0] + new_shape[0] + shift,))
        dataset[old_shape[0] + shift :] = new_data
        return

    # Determine the new shape of the dataset
    max_shape = (
        old_shape[0] + new_shape[0] + shift,
        max(old_shape[1], new_shape[1]),
        *old_shape[2:],
    )

    # Resize the dataset to the new shape
    dataset.resize(max_shape)

    # Fill the new data rows with np.nan if necessary
    if new_shape[1] < max_shape[1]:
        padded_new_data = np.full(
            (new_shape[0], max_shape[1], *old_shape[2:]), fill_value
        )
        padded_new_data[:, : new_shape[1]] = new_data
    else:
        padded_new_data = new_data

    dataset[old_shape[0] + shift :] = padded_new_data



import json
import typing as t
import warnings

import ase
import numpy as np
from ase.calculators.calculator import all_properties

from znh5md.misc import decompose_varying_shape_arrays, open_file
from znh5md.path import AttributePath, H5MDToASEMapping
from znh5md.serialization import MISSING, ORIGIN_TYPE, Frames

if t.TYPE_CHECKING:
    from znh5md.interface.io import IO


def update_frames(
    self, name: str, value: np.ndarray, origin: ORIGIN_TYPE, use_ase_calc: bool
) -> None:
    """
    Updates the specified frame data with appropriate transformations and storage logic.

    Parameters
    ----------
    name : str
        The name of the attribute to update.
    value : np.ndarray
        The array containing the new values to be set.
    origin : ORIGIN_TYPE
        The origin of the data (e.g., "calc", "info", "arrays").
    use_ase_calc : bool
        Whether to use ASE calculator for storing the data.
    """
    if name in ["positions", "numbers", "pbc", "cell"]:
        setattr(self, name, decompose_varying_shape_arrays(value, np.nan))
        return

    data = preprocess_data(value)
    store_data(self, name, data, origin, use_ase_calc)


def preprocess_data(value: np.ndarray) -> list:
    """
    Preprocess the input data by handling object, string, or numeric types.

    Parameters
    ----------
    value : np.ndarray
        The array containing the input data.

    Returns
    -------
    list
        The processed list of data.
    """
    if value.dtype.kind in ["O", "S", "U"]:
        try:
            return [json.loads(v) if v != b"" else MISSING for v in value]
        except json.JSONDecodeError:
            # compatibility for non-JSON data from other sources
            return [v.decode() if v != b"" else MISSING for v in value]
    else:
        data = decompose_varying_shape_arrays(value, np.nan)
        return [x if not np.all(np.isnan(x)) else MISSING for x in data]


def store_data(
    self, name: str, data: list, origin: ORIGIN_TYPE, use_ase_calc: bool
) -> None:
    """
    Store processed data into the appropriate attribute based on origin and conditions.

    Parameters
    ----------
    name : str
        The name of the property to store.
    data : list
        The processed data to store.
    origin : ORIGIN_TYPE
        The origin of the data (e.g., "calc", "info", "arrays").
    use_ase_calc : bool
        Whether to use ASE calculator for storing the data.

    Raises
    ------
    ValueError
        If the origin is invalid or disallowed.
    """
    if origin is not None and use_ase_calc:
        handle_origin_data(self, name, data, origin)
    else:
        assign_data_to_property(self, name, data, use_ase_calc)


def handle_origin_data(self, name: str, data: list, origin: ORIGIN_TYPE) -> None:
    """
    Handle data storage based on the specified origin.

    Parameters
    ----------
    name : str
        The name of the property to store.
    data : list
        The processed data to store.
    origin : ORIGIN_TYPE
        The origin of the data (e.g., "calc", "info", "arrays").

    Raises
    ------
    ValueError
        If the origin is invalid or disallowed.
    """
    if origin == "calc":
        self.calc[name] = data
    elif origin == "info":
        self.info[name] = data
    elif origin == "arrays":
        try:
            self.arrays[name] = np.array(data)
        except ValueError:
            warnings.warn(
                f"Could not convert data to array for '{name}'. "
                "Storing as list instead."
            )
            self.arrays[name] = data
    elif origin == "atoms":
        raise ValueError(f"Origin 'atoms' is not allowed for {name}")
    else:
        raise ValueError(f"Unknown origin: {origin}")


def assign_data_to_property(self, name: str, data: list, use_ase_calc: bool) -> None:
    """
    Assign data to the appropriate property based on its size and conditions.

    Parameters
    ----------
    name : str
        The name of the property to store.
    data : list
        The processed data to store.
    use_ase_calc : bool
        Whether to use ASE calculator for storing the data.
    """
    if name in all_properties:
        if use_ase_calc:
            self.calc[name] = data
        else:
            if isinstance(data[0], t.Sized) and len(data[0]) == len(self.numbers[0]):
                self.arrays[name] = data
            else:
                self.info[name] = data
    else:
        if isinstance(data[0], t.Sized) and len(data[0]) == len(self.numbers[0]):
            self.arrays[name] = data
        else:
            self.info[name] = data


def getitem(
    self: "IO", index: int | np.int_ | slice | np.ndarray | list[int]
) -> ase.Atoms | list[ase.Atoms]:
    """
    Retrieve frames of atoms based on the given index.

    Parameters
    ----------
    self : IO
        The IO object containing file information and settings.
    index : int, np.int_, slice, np.ndarray, or list[int]
        Index or indices specifying the frames to retrieve.

    Returns
    -------
    ase.Atoms or list[ase.Atoms]
        A single `ase.Atoms` object or a list of `ase.Atoms` objects,
        depending on the index.
    """
    frames = Frames()
    is_single_item = isinstance(index, int)
    if is_single_item:
        index = [index]

    with open_file(self.filename, self.file_handle, mode="r") as f:
        particles = f[f"/particles/{self.particles_group}"]
        process_species_group(self, frames, particles, index)
        process_particle_groups(self, frames, particles, index)

        if f"/observables/{self.particles_group}" in f:
            observables = f[f"/observables/{self.particles_group}"]
            process_observables(self, frames, observables, index)

    return list(frames) if not is_single_item else frames[0]


def process_species_group(self, frames: Frames, particles, index) -> None:
    """
    Process the 'species' group and update the frames.

    Parameters
    ----------
    self : IO
        The IO object containing file information and settings.
    frames : Frames
        The frames to update.
    particles : h5py.Group
        The particles group from the HDF5 file.
    index : list[int]
        Indices specifying the frames to retrieve.
    """
    grp = particles["species/value"]
    update_frames(
        frames, H5MDToASEMapping.species.value, grp[index], None, self.use_ase_calc
    )


def process_particle_groups(self, frames: Frames, particles, index) -> None:
    """
    Process particle groups other than 'species' and update the frames.

    Parameters
    ----------
    self : IO
        The IO object containing file information and settings.
    frames : Frames
        The frames to update.
    particles : h5py.Group
        The particles group from the HDF5 file.
    index : list[int]
        Indices specifying the frames to retrieve.
    """
    for grp_name in particles:
        if grp_name == "species":
            continue
        try:
            grp = particles[grp_name]
            origin = grp.attrs.get(AttributePath.origin.value, None)
            if grp_name == "box":
                process_box_group(self, frames, grp, index, origin)
            else:
                process_generic_group(self, frames, grp_name, grp, index, origin)
        except Exception as err:
            raise ValueError(f"Error processing group '{grp_name}'") from err


def process_box_group(self, frames: Frames, grp, index, origin) -> None:
    """
    Process the 'box' group and update the frames.

    Parameters
    ----------
    self : IO
        The IO object containing file information and settings.
    frames : Frames
        The frames to update.
    grp : h5py.Group
        The 'box' group from the HDF5 file.
    index : list[int]
        Indices specifying the frames to retrieve.
    origin : str or None
        The origin attribute of the group.
    """
    update_frames(frames, "cell", grp["edges/value"][index], origin, self.use_ase_calc)
    try:
        update_frames(frames, "pbc", grp["pbc/value"][index], origin, self.use_ase_calc)
    except KeyError:
        pbc = grp.attrs.get(AttributePath.boundary.value, ["none"] * 3)
        pbc = np.array([b != "none" for b in pbc], dtype=bool)
        frames.pbc = np.array([pbc] * len(frames))


def process_generic_group(
    self, frames: Frames, grp_name: str, grp, index, origin
) -> None:
    """
    Process generic particle groups and update the frames.

    Parameters
    ----------
    self : IO
        The IO object containing file information and settings.
    frames : Frames
        The frames to update.
    grp_name : str
        The name of the group.
    grp : h5py.Group
        The particle group from the HDF5 file.
    index : list[int]
        Indices specifying the frames to retrieve.
    origin : str or None
        The origin attribute of the group.

    Raises
    ------
    KeyError
        If the group does not contain a valid 'value' dataset.
    """
    try:
        try:
            update_frames(
                frames,
                H5MDToASEMapping[grp_name].value,
                grp["value"][index],
                origin,
                self.use_ase_calc,
            )
        except KeyError:
            update_frames(
                frames, grp_name, grp["value"][index], origin, self.use_ase_calc
            )
        except (OSError, IndexError):
            pass  # Handle backfilling for invalid values
    except KeyError:
        raise KeyError(
            f"Key '{grp_name}' does not seem to be a valid H5MD group"
            " - missing 'value' dataset."
        )


def process_observables(self, frames: Frames, observables, index) -> None:
    """
    Process observables and update the frames.

    Parameters
    ----------
    self : IO
        The IO object containing file information and settings.
    frames : Frames
        The frames to update.
    observables : h5py.Group
        The observables group from the HDF5 file.
    index : list[int]
        Indices specifying the frames to retrieve.
    """
    for grp_name in observables:
        grp = observables[grp_name]
        origin = grp.attrs.get(AttributePath.origin.value, None)
        try:
            try:
                try:
                    update_frames(
                        frames,
                        H5MDToASEMapping[grp_name].value,
                        grp["value"][index],
                        origin,
                        self.use_ase_calc,
                    )
                except KeyError:
                    update_frames(
                        frames, grp_name, grp["value"][index], origin, self.use_ase_calc
                    )
            except (OSError, IndexError):
                pass  # Handle backfilling for invalid values
        except KeyError:
            raise KeyError(
                f"Key '{grp_name}' does not seem to be a valid H5MD group"
                " - missing 'value' dataset."
            )
        except Exception as err:
            raise ValueError(f"Error processing group '{grp_name}'") from err



import typing as t

import ase
import h5py
import numpy as np

from znh5md.misc import fill_dataset, open_file
from znh5md.path import AttributePath, get_h5md_path
from znh5md.serialization import Entry, Frames

if t.TYPE_CHECKING:
    from znh5md.interface.io import IO


def create_group(
    f,
    path,
    entry: Entry,
    ref_length: int,
    particles_goup: str,
    pbc_grp: bool,
    store: t.Literal["time", "linear"],
    save_units: bool,
    timestep: float,
    compression: str | None,
    compression_opts: int | None,
    chunk_size: int | None,
) -> None:
    if path in f:
        raise ValueError(f"Group {path} already exists")
    data, dtype = entry.dump()

    if entry.name == "pbc":
        box_grp = f[f"/particles/{particles_goup}"].create_group("box")
        box_grp.attrs.create(
            AttributePath.boundary.value,
            ["periodic" if x else "none" for x in entry.ref],
        )
        box_grp.attrs.create(AttributePath.dimension.value, 3)
        if not pbc_grp:
            return

    grp = f.create_group(path)

    if dtype == h5py.string_dtype():
        chunks = True if chunk_size is None else (chunk_size,)
        ds = grp.create_dataset(
            "value",
            shape=(ref_length + len(data),),
            maxshape=(None,),
            fillvalue=entry.fillvalue,
            dtype=dtype,
            compression=compression,
            compression_opts=compression_opts,
            chunks=chunks,
        )
        ds[ref_length:] = data
    else:
        maxshape = tuple(None for _ in data.shape)
        shape = (ref_length + len(data),) + data.shape[1:]
        chunks = (
            True if chunk_size is None else tuple([chunk_size] + list(data.shape[1:]))
        )

        ds = grp.create_dataset(
            "value",
            shape=shape,
            maxshape=maxshape,
            fillvalue=entry.fillvalue,
            dtype=dtype,
            compression=compression,
            compression_opts=compression_opts,
            chunks=chunks,
        )
        ds[ref_length:] = data

    if entry.origin is not None:
        grp.attrs.create(AttributePath.origin.value, entry.origin)
    if entry.unit is not None and save_units:
        ds.attrs.create(AttributePath.unit.value, entry.unit)

    # We use linear time and step for now
    # because most of the time we don't have an step / offset ...
    if store == "time":
        grp.create_dataset("step", data=np.arange(1, len(ds) + 1), maxshape=(None,))
    else:
        grp.create_dataset("step", data=1)
    # step_ds.attrs.create(AttributePath.unit.value, "fs")

    if store == "time":
        time_ds = grp.create_dataset(
            "time", data=np.arange(1, len(ds) + 1) * timestep, maxshape=(None,)
        )
    else:
        time_ds = grp.create_dataset("time", data=timestep)
    time_ds.attrs.create(AttributePath.unit.value, "fs")


def extend_group(
    f,
    path,
    entry: Entry,
    ref_length: int,
    store: t.Literal["time", "linear"],
    timestep: float,
) -> None:
    if path not in f:
        raise ValueError(f"Group {path} not found exists")

    grp = f[path]
    data, dtype = entry.dump()

    shift = ref_length - len(grp["value"])
    if dtype == h5py.string_dtype():
        grp["value"].resize((len(grp["value"]) + len(data) + shift,))
        grp["value"][len(grp["value"]) - len(data) :] = data
    else:
        fill_dataset(grp["value"], data, shift, entry.fillvalue)

    if store == "time":
        step_ds = grp["step"]
        step_ds.resize((ref_length,))
        step_ds[:] = np.arange(1, ref_length + 1)
        time_ds = grp["time"]
        time_ds.resize((ref_length,))
        time_ds[:] = np.arange(1, ref_length + 1) * timestep


def extend(self: "IO", data: list[ase.Atoms]) -> None:
    if self.particles_group is None:
        raise ValueError("Particles group not set")

    # TODO: flag to save with origin=None to test against files no coming from znh5md

    frames = Frames.from_ase(data)
    frames.check()

    species_path = get_h5md_path("numbers", self.particles_group, frames)

    with open_file(self.filename, self.file_handle, mode="a") as f:
        if species_path in f:
            ref_length = len(f[species_path]["value"])
        else:
            ref_length = 0

        for entry in frames.items():
            path = get_h5md_path(entry.name, self.particles_group, frames)
            if path in f:
                extend_group(
                    f, path, entry, ref_length, store=self.store, timestep=self.timestep
                )
            else:
                if not self._store_ase_origin:
                    entry.origin = None
                # TODO: what about extending with pbc group false?
                create_group(
                    f,
                    path,
                    entry,
                    ref_length,
                    self.particles_group,
                    pbc_grp=self.pbc_group,
                    store=self.store,
                    save_units=self.save_units,
                    timestep=self.timestep,
                    compression=self.compression,
                    compression_opts=self.compression_opts,
                    chunk_size=self.chunk_size,
                )



from enum import Enum


class H5MDUnits(str, Enum):
    energy = "eV"
    forces = "eV/Angstrom"
    stress = "eV/Angstrom^3"
    velocities = "Angstrom/fs"
    positions = "Angstrom"
    time = "fs"
    cell = "Angstrom"


def get_unit(key: str) -> str | None:
    try:
        return H5MDUnits[key].value
    except KeyError:
        return None



import pathlib
import typing as t

import ase

from znh5md.interface import IO


def read(filename, index: int | slice | list[int] = -1) -> ase.Atoms | list[ase.Atoms]:
    io = IO(filename)
    return io[index]


def write(
    filename, images: ase.Atoms | list[ase.Atoms], append: bool = True, **kwargs
) -> None:
    if not append and pathlib.Path(filename).exists():
        raise FileExistsError(
            f"{filename} already exists. Remove it or set append=True."
        )
    if isinstance(images, ase.Atoms):
        images = [images]
    io = IO(filename, **kwargs)
    io.extend(images)


def iread(filename) -> t.Iterable[ase.Atoms]:
    io = IO(filename)
    for atoms in io:
        yield atoms
